"""Implementation of various pytorch models."""
from swem.models.pooling import (
    AttentionPooling,
    HierarchicalPooling,
    MaxPooling,
    MeanPooling,
    SwemPoolingLayer,
)
from swem.models.swem import Swem
from swem.models.word_drop_embedding import WordDropEmbedding



"""Implementation of certain useful metrics."""
from typing import Dict, List, Optional, Union

import torch

from swem.utils.classes import ClfMetricTracker, KeyDependentDefaultdict


class ClassificationReport:
    """A class for tracking various metrics in a classification task.

    The class is particularly useful when doing handling a dataset batch-wise where
    metrics should be aggregated over the whole dataset.

    Args:
        target_names (Optional[List[str, int]]): Labels in the classification task in
          the same order as the output of the model.
        binary (bool): Whether or not we are doing binary classification (i.e. the
          model output is the pre-sigmoid logit for the positive class). Defaults to
          False.

    Examples:
        >>> report = ClassificationReport(target_names=["A", "B"])
        >>> logits = torch.tensor([[1,0], [0,1], [1,0]])
        >>> labels = torch.tensor([0, 0, 1])
        >>> report.update(logits, labels)
        >>> report.get()
        {
            "num_samples": 3,
            "accuracy": 0.3333333333333333,
            "class_metrics": {
                "A": {
                "support": 2,
                "recall": 0.5,
                "precision": 0.5,
                "f1_score": 0.5
                },
                "B": {
                "support": 1,
                "recall": 0.0,
                "precision": 0.0,
                "f1_score": null
                }
            }
        }
        >>> mask = torch.tensor([1,1,0])
        >>> report.reset()
        >>> report.update(logits, labels, mask)
        >>> report.get()
        {
            "num_samples": 2,
            "accuracy": 0.5,
            "class_metrics": {
                "A": {
                "support": 2,
                "recall": 0.5,
                "precision": 1.0,
                "f1_score": 0.6666666666666666
                },
                "B": {
                "support": 0,
                "recall": null,
                "precision": 0.0,
                "f1_score": null
                }
            }
        }
    """

    def __init__(
        self, target_names: Optional[List[Union[str, int]]] = None, binary: bool = False
    ):
        self.target_names = target_names
        self.binary = binary
        self.num_samples = 0
        self.num_correct = 0
        self.class_metrics = KeyDependentDefaultdict(ClfMetricTracker)

    def reset(self):
        """Reset all tracked values."""
        self.num_samples = 0
        self.num_correct = 0
        self.class_metrics = KeyDependentDefaultdict(ClfMetricTracker)

    def get(
        self,
    ) -> Optional[
        Dict[str, Union[int, float, Dict[Union[int, str], Dict[str, Optional[float]]]]]
    ]:
        """Get the current state of all tracked metrics."""
        if self.num_samples == 0:
            return None
        return {
            "num_samples": self.num_samples,
            "accuracy": self.num_correct / self.num_samples,
            "class_metrics": {
                name: {
                    "support": met.support,
                    "recall": met.recall,
                    "precision": met.precision,
                    "f1_score": met.f1_score,
                }
                for name, met in self.class_metrics.items()
            },
        }

    def update(
        self,
        logits: torch.Tensor,
        labels: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
    ):
        """Update the tracked metrics with the results from a new batch.

        Args:
            logits (torch.Tensor): Output of the model in the classification task
              (pre-softmax/sigmoid).
            labels (torch.Tensor): The correct labels.
            mask (Optional[torch.Tensor], optional): A 0/1-mask telling us which
              samples to take into account (where mask is 1). Defaults to None.
        """
        if self.binary:
            num_classes = 2
            preds = torch.tensor(logits > 0).to(dtype=torch.int64).view(-1)
        else:
            num_classes = logits.size(-1)
            preds = torch.argmax(torch.tensor(logits), dim=-1).view(-1)

        labels = labels.view(-1)
        if mask is None:
            mask = torch.ones_like(labels, dtype=torch.bool, device=labels.device)
        else:
            mask = torch.tensor(mask, dtype=torch.bool)

        assert labels.size() == mask.size()
        assert preds.size() == mask.size()

        self.num_samples += torch.sum(mask).item()
        self.num_correct += torch.sum((labels == preds) * mask).item()

        for i in range(num_classes):
            if self.target_names is None:
                name = i
            else:
                name = self.target_names[i]

            label_mask = labels == i
            preds_mask = preds == i

            self.class_metrics[name].tp += torch.sum(
                label_mask * preds_mask * mask
            ).item()
            self.class_metrics[name].fp += torch.sum(
                (~label_mask) * preds_mask * mask
            ).item()
            self.class_metrics[name].fn += torch.sum(
                label_mask * (~preds_mask) * mask
            ).item()



"""A collection of pooling layers with a common API."""
from typing import Dict, Optional, Union

import torch
from torch import nn


class SwemPoolingLayer(nn.Module):
    """Base class for all pooling layers."""

    def forward(
        self, input: torch.FloatTensor, mask: Optional[torch.FloatTensor] = None
    ) -> torch.FloatTensor:
        """The pooling computation. This should be overridden by subclasses."""
        raise NotImplementedError()

    @staticmethod
    def from_config(config: Dict[str, Union[str, int]]) -> "SwemPoolingLayer":
        """Construct a pooling layer from the config.

        The config should have a key 'class' specifying the pooling layer to construct,
        e.g. 'AttentionPooling' and additional keys providing the arguments for this
        class's __init__ (e.g. for AttentionPooling 'input_dim').

        Raises:
            NotImplementedError: If the class specified by 'class' is unknown.

        Returns:
            SwemPoolingLayer: A pooling layer defined by the config.
        """
        pooling_type = config["class"]
        if pooling_type == "HierarchicalPooling":
            return HierarchicalPooling(window_size=config["window_size"])
        elif pooling_type == "MaxPooling":
            return MaxPooling()
        elif pooling_type == "MeanPooling":
            return MeanPooling()
        elif pooling_type == "AttentionPooling":
            return AttentionPooling(input_dim=config["input_dim"])
        else:
            raise NotImplementedError(f"Unknown class {pooling_type}")


class HierarchicalPooling(SwemPoolingLayer):
    """Hierarchical Pooling layer (see
    `Baselines need more love <https://arxiv.org/abs/1808.09843>`_ ).

    First mean pooling along the sequence dimension over windows of the given size,
    then max pooling along the sequence dimension.

    The mask input is ignored; it is only there for compatibility.

    Args:
        window_size (int): Size of the pooling window in the mean pooling step.


    Shapes:
        - input: :math:`(\\text{batch_size}, \\text{seq_len}, \\text{enc_dim})`
        - mask: :math:`(\\text{batch_size}, \\text{enc_dim})`
        - output: :math:`(\\text{batch_size}, \\text{enc_dim})`
    """

    def __init__(self, window_size: int):
        super().__init__()
        self.avg_pooling = nn.AvgPool2d(kernel_size=(window_size, 1), stride=1)
        self.window_size = window_size

    @property
    def config(self) -> Dict[str, Union[str, int]]:
        return {"class": "HierarchicalPooling", "window_size": self.window_size}

    def forward(
        self, input: torch.FloatTensor, mask: Optional[torch.FloatTensor] = None
    ) -> torch.FloatTensor:
        """Pooling forward pass."""

        seq_len = input.size(1)

        # If the sequenze is particularly short there is not much to do
        if seq_len <= self.window_size:
            return torch.mean(input, dim=1)

        output = self.avg_pooling(input.unsqueeze(1)).squeeze(1)
        output, _ = torch.max(output, dim=-2)
        return output


class MeanPooling(SwemPoolingLayer):
    """Simple mean pooling layer.

    Mean pooling along the sequence dimension; ignoring inputs according to the given
    mask.


    Shapes:
        - input: :math:`(\\text{batch_size}, \\text{seq_len}, \\text{enc_dim})`
        - mask: :math:`(\\text{batch_size}, \\text{enc_dim})`
        - output: :math:`(\\text{batch_size}, \\text{enc_dim})`
    """

    @property
    def config(self) -> Dict[str, str]:
        return {"class": "MeanPooling"}

    def forward(
        self, input: torch.FloatTensor, mask: Optional[torch.FloatTensor] = None
    ) -> torch.FloatTensor:
        """Pooling forward pass."""

        if mask is None:
            return torch.mean(input, dim=-2)
        else:
            masked_input = input * mask.unsqueeze(-1)
            output = torch.sum(masked_input, dim=-2) / torch.sum(
                mask, dim=-1, keepdim=True
            )
            return output


class MaxPooling(SwemPoolingLayer):
    """Simple max pooling layer.

    Max pooling along the sequence dimension.

    The mask input is ignored; it is only there for compatibility.


    Shapes:
        - input: :math:`(\\text{batch_size}, \\text{seq_len}, \\text{enc_dim})`
        - mask: :math:`(\\text{batch_size}, \\text{enc_dim})`
        - output: :math:`(\\text{batch_size}, \\text{enc_dim})`
    """

    @property
    def config(self) -> Dict[str, str]:
        return {"class": "MaxPooling"}

    def forward(
        self, input: torch.FloatTensor, mask: Optional[torch.FloatTensor] = None
    ) -> torch.FloatTensor:
        """Pooling forward pass."""

        output, _ = torch.max(input, dim=-2)
        return output


class AttentionPooling(SwemPoolingLayer):
    """Pooling with an attention mechanism.

    Pools along the sequence dimension by computing a weighted sum whose weights
    are computed as a small feed forward network (with output size 1) applied to the
    input vectors, followed by a softmax.

    Takes an optional mask telling us which inputs to ignore for the softmax.

    Args:
        input_dim (int): The size of the input vectors.

    Shapes:
        - input: :math:`(\\text{batch_size}, \\text{seq_len}, \\text{enc_dim})`
        - mask: :math:`(\\text{batch_size}, \\text{enc_dim})`
        - output: :math:`(\\text{batch_size}, \\text{enc_dim})`
    """

    def __init__(self, input_dim: int):
        super().__init__()
        self.attention_trafo = nn.Sequential(
            nn.Linear(input_dim, input_dim),
            nn.ReLU(),
            nn.Linear(input_dim, 1, bias=False),
        )

    @property
    def config(self) -> Dict[str, Union[int, str]]:
        return {
            "class": "AttentionPooling",
            "input_dim": self.attention_trafo[0].in_features,
        }

    def forward(
        self, input: torch.FloatTensor, mask: Optional[torch.FloatTensor] = None
    ) -> torch.FloatTensor:
        """Pooling forward pass."""
        attention_logits = self.attention_trafo(input).squeeze(-1)
        attention_weights = torch.softmax(attention_logits, dim=-1)
        if mask is not None:
            # If we have a mask we null the corresponding attention weights
            # and then renormalize (this has the same effect as only doing
            # the softmax over the non-masked inputs)
            attention_weights = attention_weights * mask
            attention_weights = attention_weights / torch.sum(
                attention_weights, dim=-1, keepdim=True
            )

        return torch.einsum("bsd,bs->bd", input, attention_weights)



"""Implementation of the Simple Word Embedding Modell."""

from itertools import chain
from typing import Any, Dict, Optional, Tuple

import torch
from torch import nn

from swem.models.pooling import SwemPoolingLayer
from swem.models.word_drop_embedding import WordDropEmbedding


class Swem(nn.Module):
    """Simple Word Embedding model (see
    `Baselines need more love <https://arxiv.org/abs/1808.09843>`_ ).

    The model consists of an embedding layer, a feed forward network that is applied
    separately to each word vector, a pooling layer that pools the vectors belonging to
    the same text into a single vector, and another feed forward network that is applied
    to this pooled vector.

    Args:
        embedding (nn.Embedding): The embedding layer used by the model.
        pooling_layer (nn.Module): The pooling layer to be used by the model.
        pre_pooling_dims (Optional[Tuple[int, ...]]): Intermediate dimensions for the
          feed forward network applied to the input.
        post_pooling_dims (Optional[Tuple[int, ...]]): Intermediate dimensions for the
          feed forward network applied to the output of the pooling layer.
        dropout (float): Dropout probability after each layer in both feed forward
          subnetworks.

    Shapes:
        - input: :math:`(\\text{batch_size}, \\text{seq_len})`
        - output: :math:`(\\text{batch_size}, \\text{enc_dim})`, where
          :math:`\\text{enc_dim}` is the last of the post_pooling_dims (if given,
          otherwise the last of the pre_pooling_dims or failing that the
          embedding_dimension).

    """

    def __init__(
        self,
        embedding: nn.Embedding,
        pooling_layer: SwemPoolingLayer,
        pre_pooling_dims: Optional[Tuple[int, ...]] = None,
        post_pooling_dims: Optional[Tuple[int, ...]] = None,
        dropout: float = 0.2,
    ):
        super().__init__()
        self.embedding = embedding
        self.pooling_layer = pooling_layer
        self.pre_pooling_dims = pre_pooling_dims
        self.post_pooling_dims = post_pooling_dims
        self.dropout = dropout

        if pre_pooling_dims is None:
            self.pre_pooling_trafo: nn.Module = nn.Identity()
        else:
            pre_dims = [embedding.embedding_dim, *pre_pooling_dims]
            self.pre_pooling_trafo = nn.Sequential(
                *chain(
                    *[
                        (nn.Linear(dim_in, dim_out), nn.ReLU(), nn.Dropout(dropout))
                        for dim_in, dim_out in zip(pre_dims[:-1], pre_dims[1:])
                    ]
                )
            )

        if post_pooling_dims is None:
            self.post_pooling_trafo: nn.Module = nn.Identity()
        else:
            pooling_dim = (
                embedding.embedding_dim
                if pre_pooling_dims is None
                else pre_pooling_dims[-1]
            )
            if len(post_pooling_dims) == 1:
                self.post_pooling_trafo = nn.Linear(pooling_dim, post_pooling_dims[0])
            else:
                post_dims = [pooling_dim, *post_pooling_dims[:-1]]
                self.post_pooling_trafo = nn.Sequential(
                    *chain(
                        *[
                            (nn.Linear(dim_in, dim_out), nn.ReLU(), nn.Dropout(dropout))
                            for dim_in, dim_out in zip(post_dims[:-1], post_dims[1:])
                        ]
                    ),
                    torch.nn.Linear(post_dims[-1], post_pooling_dims[-1]),
                )

    @property
    def config(self) -> Dict[str, Any]:
        embedding_config = {
            "class": "WordDropEmbedding"
            if isinstance(self.embedding, WordDropEmbedding)
            else "Embedding",
            "embedding_dim": self.embedding.embedding_dim,
            "num_embeddings": self.embedding.num_embeddings,
            "padding_idx": self.embedding.padding_idx,
            "scale_grad_by_freq": self.embedding.scale_grad_by_freq,
            "max_norm": self.embedding.max_norm,
            "norm_type": self.embedding.norm_type,
            "sparse": self.embedding.sparse,
        }

        if isinstance(self.embedding, WordDropEmbedding):
            embedding_config["p"] = self.embedding.p

        return {
            "pre_pooling_dims": self.pre_pooling_dims,
            "post_pooling_dims": self.post_pooling_dims,
            "dropout": self.dropout,
            "pooling": self.pooling_layer.config,
            "embedding": embedding_config,
        }

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "Swem":
        """Construct a SWEM-model from a config.

        The config should have the following keys:

            - pooling: A config for a SwemPoolingLayer.
            - embedding: A dict with key 'class' specifying either 'Embedding' or
              'WordDropEmbedding' and further keys according to the arguments of the
              correct class's __init__.
            - pre_pooling_dims: Tuple of int as for __init__.
            - post_pooling_dims: Tuple of int as for __init__.
            - dropout: float as for __init__.

        Args:
            config (Dict[str, Any]): The config to construct the model from.

        Examples:
            >>> config = {
            ...     "embedding": {
            ...             "class": "Embedding",
            ...             "num_embeddings": 10,
            ...             "embedding_dim": 2
            ...     },
            ...     "pooling": {
            ...             "class": "HierarchicalPooling",
            ...             "window_size": 5
            ...     },
            ...     "pre_pooling_dims": (5, 5),
            ...     "post_pooling_dims": (6, 6),
            ...     "dropout": 0.1
            ... }
            >>> Swem.from_config(config)
            Swem(
            (embedding): Embedding(10, 2)
            (pooling_layer): HierarchicalPooling(
                (avg_pooling): AvgPool2d(kernel_size=(5, 1), stride=1, padding=0)
            )
            (pre_pooling_trafo): Sequential(
                (0): Linear(in_features=2, out_features=5, bias=True)
                (1): ReLU()
                (2): Dropout(p=0.1, inplace=False)
                (3): Linear(in_features=5, out_features=5, bias=True)
                (4): ReLU()
                (5): Dropout(p=0.1, inplace=False)
            )
            (post_pooling_trafo): Sequential(
                (0): Linear(in_features=5, out_features=6, bias=True)
                (1): ReLU()
                (2): Dropout(p=0.1, inplace=False)
                (3): Linear(in_features=6, out_features=6, bias=True)
            )
            )

        """
        pooling_config = config.pop("pooling")
        pooling_layer = SwemPoolingLayer.from_config(pooling_config)

        embedding_config = config.pop("embedding")
        if embedding_config.pop("class") == "WordDropEmbedding":
            embedding = WordDropEmbedding(**embedding_config)
        else:
            embedding = nn.Embedding(**embedding_config)

        return cls(embedding=embedding, pooling_layer=pooling_layer, **config)

    def forward(self, input: torch.Tensor) -> torch.FloatTensor:
        output = self.embedding(input)
        output = self.pre_pooling_trafo(output)
        output = self.pooling_layer(output)
        output = self.post_pooling_trafo(output)
        return output



"""Some useful helper classes."""

from dataclasses import dataclass
from typing import Optional, Union


class KeyDependentDefaultdict(dict):
    """Defaultdict where the default factory can depend on the key.

    Args:
        default_factory (callable): Default factory called when a missing key is
          encountered. The call will be default_factory(key) so the callable should
          take exactly one argument.

    Examples:
        >>> d = KeyDependentDefaultdict(lambda key: {"name": key})
        >>> d["a"]["b"] = 1
        >>> d
        {'a': {'name': 'a', 'b': 1}}
    """

    def __init__(self, /, *args, **kwargs):
        if not callable(args[0]):
            raise ValueError("First argument must be a callable.")

        self.default_factory = args[0]
        super().__init__(*args[1:], **kwargs)

    def __missing__(self, key):
        res = self.default_factory(key)
        self[key] = res
        return res


@dataclass
class ClfMetricTracker:
    """Class for keeping track of the metrics corresponding to a single label in a
    classification task.

    Args:
        name: The label this instance is keeping track of.
        tp: Start value for true positives. Defaults to 0.
        fp: Start-value for false positives. Defaults to 0.
        fn: Start-value for false negatives. Defaults to 0.

    Attributes:
        support: Number of instances for the label that were encountered.
        recall: Current value for the recall metric.
        precision: Current value for the precision metric.
        f1_score: Current value for the f1_score metric.

    Examples:
        >>> from swem.utils.classes import ClfMetricTracker
        >>> tracker = ClfMetricTracker(name="Label_1", tp=5, fp=1, fn=3)
        >>> tracker
        ClfMetricTracker(name='Label_1', tp=5, fp=1, fn=3)
        >>> tracker.support
        8
        >>> tracker.recall
        0.625
        >>> tracker.precision
        0.8333333333333334
        >>> tracker.f1_score
        0.7142857142857143

    """

    name: Optional[Union[str, int]] = None
    tp: int = 0
    fp: int = 0
    fn: int = 0

    @property
    def support(self) -> int:
        return self.tp + self.fn

    @property
    def recall(self) -> Optional[float]:
        if self.support == 0:
            return None
        return self.tp / self.support

    @property
    def precision(self) -> Optional[float]:
        if self.tp + self.fp == 0:
            return None
        return self.tp / (self.tp + self.fp)

    @property
    def f1_score(self):
        if not self.recall or not self.precision:
            return None
        return 2 * self.recall * self.precision / (self.recall + self.precision)



"""Embedding layer with word-drop regularization."""
import torch
from torch import nn


class WordDropEmbedding(nn.Embedding):
    """Embedding layer with word-drop regularization.

    During training drops certain words (token ids) entirely from batches (zeroing
    the corresponding vectors). This layer can be used as a drop-in replacement for
    the usual nn.Embedding.

    Args:
        p (float): Probability with which to drop words (if 0 this layer behaves
          just like a usual embedding layer).
        args/kwargs: Other args and kwargs are handled by nn.Embedding.

    Examples::

        >>> emb = WordDropEmbedding(10, 2, p=0.3)
        >>> input = torch.arange(0, 10, dtype=torch.int64)
        >>> output = emb(input)
        >>> print(output)
        tensor([[-0.8393,  1.3216],
                [-0.3652,  0.2879],
                [ 0.1899,  2.2358],
                [ 1.7776, -1.8437],
                [-0.6406,  0.7939],
                [ 1.0874, -3.2290],
                [ 0.0000, -0.0000],
                [ 0.0000,  0.0000],
                [-0.9666,  1.0100],
                [-0.2444, -2.2618]], grad_fn=<DivBackward0>)

    """

    def __init__(self, *args, p: float, **kwargs):
        super().__init__(*args, **kwargs)
        if p < 0 or p >= 1:
            raise ValueError(
                f"Dropout probability must be non-negative and less than 1, got {p}"
            )
        self.p = p

    def forward(self, input: torch.Tensor) -> torch.FloatTensor:
        output = super().forward(input)
        if self.training and self.p > 0:
            # Compute a mask telling us which ids in the vocab to drop
            vocab_mask = torch.bernoulli(
                torch.Tensor().new_full(
                    (self.num_embeddings,), 1 - self.p, device=input.device
                )
            )

            # The corresponding mask for the input
            batch_mask = torch.index_select(vocab_mask, 0, input.view(-1)).view(
                *input.size(), 1
            )

            # Null the embedded vectors and rescale
            output = (output * batch_mask) / (1 - self.p)

        return output



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="pytorch-swem",
    version="0.3.0",
    author="Sebastian Schönnenbeck",
    author_email="schoennenbeck@gmail.com",
    url="https://github.com/schoennenbeck/swem",
    project_urls={"Documentation": "https://pytorch-swem.readthedocs.io/"},
    description="Pytorch implementation of the simple word embedding model.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["tests"]),
    install_requires=["torch >= 1.7.1"],
    license="MIT",
    python_requires=">=3.8",
    extras_require={
        "dev": [
            "black",
            "isort",
            "mypy",
            "pre-commit",
            "pytest",
            "pytest-cov",
            "coveralls",
            "coverage",
        ],
        "docs": [
            "sphinx",
            "myst-parser",
            "sphinx-rtd-theme",
            "sphinx-autodoc-typehints",
        ],
    },
)
